#include "Params.h"

Params::Params(
	float* nodes_coords, 
	float *demands, 
	int nodes_num,
	int corrds_scale,
	int demands_scale,
	int seed
){
	double serviceTimeData = 0.;
	nbClients = nodes_num - 1;
	totalDemand = 0.;
	maxDemand = 0.;
	durationLimit = 1.e30;
	vehicleCapacity = (float)(demands_scale);
	isRoundingInteger = true;
	isDurationConstraint = false;

	// Initialize RNG
	srand(seed);	

	// Reading client coordinates
	cli = std::vector<Client>(nbClients + 1);
	for (int i = 0; i <= nbClients; i++)
	{
		cli[i].custNum =i;
		cli[i].coordX = nodes_coords[2*i] * (float)(corrds_scale);
		cli[i].coordY = nodes_coords[2*i+1] * (float)(corrds_scale);
		cli[i].polarAngle = CircleSector::positive_mod(
			32768.*atan2(cli[i].coordY - cli[0].coordY, cli[i].coordX - cli[0].coordX) / PI
		);
	}

	// Reading demand information
	for (int i = 0; i <= nbClients; i++)
	{
		cli[i].demand = demands[i] * vehicleCapacity;
		cli[i].serviceDuration = (i == 0) ? 0. : serviceTimeData ;
		if (cli[i].demand > maxDemand) maxDemand = cli[i].demand;
		totalDemand += cli[i].demand;
	}
	
	// Default initialization if the number of vehicles has not been provided by the user
	nbVehicles = std::ceil(1.2 * totalDemand / vehicleCapacity) + 2;  // Safety margin: 20% + 2 more vehicles than the trivial bin packing LB

	// Calculation of the distance matrix
	maxDist = 0.;
	timeCost = std::vector < std::vector< double > >(nbClients + 1, std::vector <double>(nbClients + 1));
	for (int i = 0; i <= nbClients; i++)
	{
		for (int j = 0; j <= nbClients; j++)
		{
			double d = std::sqrt((cli[i].coordX - cli[j].coordX)*(cli[i].coordX - cli[j].coordX) + (cli[i].coordY - cli[j].coordY)*(cli[i].coordY - cli[j].coordY));
			if (isRoundingInteger) { d += 0.5; d = (double)(int)d; } // integer rounding
			if (d > maxDist) maxDist = d;
			timeCost[i][j] = d;
		}
	}

	// Calculation of the correlated vertices for each customer (for the granular restriction)
	correlatedVertices = std::vector < std::vector < int > >(nbClients + 1);
	std::vector < std::set < int > > setCorrelatedVertices = std::vector < std::set <int> >(nbClients + 1);
	std::vector < std::pair <double, int> > orderProximity;
	for (int i = 1; i <= nbClients; i++)
	{
		orderProximity.clear();
		for (int j = 1; j <= nbClients; j++)
			if (i != j) orderProximity.push_back(std::pair <double, int>(timeCost[i][j], j));
		std::sort(orderProximity.begin(), orderProximity.end());

		for (int j = 0; j < std::min<int>(nbGranular, nbClients - 1); j++)
		{
			// If i is correlated with j, then j should be correlated with i
			setCorrelatedVertices[i].insert(orderProximity[j].second);
			setCorrelatedVertices[orderProximity[j].second].insert(i);
		}
	}

	// Filling the vector of correlated vertices
	for (int i = 1; i <= nbClients; i++)
		for (int x : setCorrelatedVertices[i])
			correlatedVertices[i].push_back(x);

	// Safeguards to avoid possible numerical instability in case of instances containing arbitrarily small or large numerical values
	if (maxDist < 0.1 || maxDist > 100000)   throw std::string("The distances are of very small or large scale. This could impact numerical stability. Please rescale the dataset and run again.");
	if (maxDemand < 0.1 || maxDemand > 100000) throw std::string("The demand quantities are of very small or large scale. This could impact numerical stability. Please rescale the dataset and run again.");
	if (nbVehicles < std::ceil(totalDemand / vehicleCapacity)) throw std::string("Fleet size is insufficient to service the considered clients.");

	// A reasonable scale for the initial values of the penalties
	penaltyDuration = 1;
	penaltyCapacity = std::max<double>(0.1, std::min<double>(1000., maxDist / maxDemand));
}